#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include "main.h"
#include "global.h"
#include "init.h"

void get_options(int argc, char *argv[]) {
/***********************************************************************
Get the command line options.  The results are stored in global
variables defined in global.h .
***********************************************************************/
  int n;

  /* Default values. */
  CallFileName = strdup("in");
  LogFileName = NULL;
  LoggingFlag = FALSE;
  SummaryFileName = NULL;
  SummaryFlag = FALSE;
  NTest = 1000;
  MaxParameters = MAX_PARAMETERS;
  Rand48State[0] = generate_seed();
  Rand48State[1] = Rand48State[0];
  Rand48State[2] = Rand48State[0];
  AggressiveFlag = FALSE;
  PrintEachTestFlag = FALSE;
  AnnounceTestsFlag = FALSE;
  SignalParent = FALSE;

  /*********************************************************************
  Save the information for each specified option.
  *********************************************************************/

  for (n = 1; n < argc; n++) {

    if (argv[n][0] != '-') {
      fprintf(stderr, "Error:  Argument '%s' is not valid.\n", argv[n]);
      usage(argv[0]);
      exit(1);
    }

#   ifdef ECHO_OPTIONS
      printf("Option '%s':", &argv[n][1]);
#   endif

    if (!strcmp(&argv[n][1], "callfile")) {
    /*******************************************************************
    The file with the list of system calls to be fuzzed.
    *******************************************************************/
      CallFileName = argv[++n];
#     ifdef ECHO_OPTIONS
        printf("  The call file is '%s'.\n", CallFileName);
#     endif

    } else if (!strcmp(&argv[n][1], "tests")) {
    /*******************************************************************
    The number of tests to perform.
    *******************************************************************/
      NTest = atoi(&argv[++n][0]);
#     ifdef ECHO_OPTIONS
        printf("  Performing %d tests.\n", NTest);
#     endif

    } else if (!strcmp(&argv[n][1], "max_parms")) {
    /*******************************************************************
    Maximum number of parameters allowed for a system call.
    *******************************************************************/
      MaxParameters = atoi(&argv[++n][0]);
#     ifdef ECHO_OPTIONS
        printf("  Maximum allowed parameters for a call is %d.\n",
	       MaxParameters);
#     endif
      assert(MaxParameters <= MAX_PARAMETERS);

    } else if (!strcmp(&argv[n][1], "logfile")) {
    /*******************************************************************
    The file to which the log is written.  If the file is given, we
    assume logging.  If it is not given, we assume no logging.
    *******************************************************************/
      LogFileName = argv[++n];
#     ifdef ECHO_OPTIONS
        printf("  The log file is '%s'.\n", LogFileName);
#     endif
      LoggingFlag = TRUE;

    } else if (!strcmp(&argv[n][1], "summaryfile")) {
    /*******************************************************************
    The file to which the summary is written.  When the run starts,
    information about this run is appended to the summary file.  This is
    useful when running the program from several starting points with no
    logging.  If it is not given, we assume no summary file is to be
    written.
    *******************************************************************/
      SummaryFileName = argv[++n];
#     ifdef ECHO_OPTIONS
        printf("  The summary file is '%s'.\n", SummaryFileName);
#     endif
      SummaryFlag = TRUE;

    } else if (!strcmp(&argv[n][1], "seed")) {
    /*******************************************************************
    The initial state for the random number generator, rand48().

    This parameter expects only one integer value to follow it.  It does
    NOT do robust error checking.  Sets the random_state to be this single 
    number repeated 3 times.  
    *******************************************************************/
      Rand48State[0] = atoi(&argv[++n][0]);
      Rand48State[1] = Rand48State[0];
      Rand48State[2] = Rand48State[0];
#     ifdef ECHO_OPTIONS
        printf("  Initial seed:  %d.\n", Rand48State[0]);
#     endif

    } else if (!strcmp(&argv[n][1], "random_state")) {
    /*******************************************************************
    The initial state for the random number generator, rand48().

    This parameter expects three integer values to follow it.  It does
    NOT do robust error checking.  This allows more possible states than
    simply using the "seed" option.
    *******************************************************************/
      Rand48State[0] = atoi(&argv[++n][0]);
      Rand48State[1] = atoi(&argv[++n][0]);
      Rand48State[2] = atoi(&argv[++n][0]);
#     ifdef ECHO_OPTIONS
        printf("  Initial state:  %d %d %d.\n",
	       Rand48State[0], Rand48State[1], Rand48State[2]);
#     endif

    } else if (!strcmp(&argv[n][1], "aggressive")) {
    /*******************************************************************
    This parameter turns on aggressive system call testing.  This means
    that the tester treats all parameters for the system calls as 
    pointers.  
    WARNING: This option will generate lots of garbage files in the 
    directory the tester is run.  Some are hard to get rid of.  Use at 
    your own risk.
    *******************************************************************/
#     ifdef ECHO_OPTIONS
        printf("  Doing aggressive testing.\n");
#     endif
      AggressiveFlag = TRUE;

    } else if (!strcmp(&argv[n][1], "catch_sigsegv")) {
    /*******************************************************************
    If this parameter is turned on, any occurrence of a segmentation
    fault is caught and ignored.  This is most useful for testing HPUX,
    as it appears that passing a bad pointer to a system call in that
    kernel causes a signal to the user code, rather than returning an
    error.
    *******************************************************************/
#     ifdef ECHO_OPTIONS
        printf("  Catching and ignoring the SIGSEGV signal.\n");
#     endif
      CatchSIGSEGVFlag = TRUE;

    } else if (!strcmp(&argv[n][1], "print_tests")) {
    /*******************************************************************
    This parameter causes information about each test of a system call
    to be sent to stdout.  This is useful for identifying calls that
    cause a system hang when first testing a new system.
    *******************************************************************/
#     ifdef ECHO_OPTIONS
        printf("  Printing information about each system call test.\n");
#     endif
      PrintEachTestFlag = TRUE;

    } else if (!strcmp(&argv[n][1], "announce_tests")) {
    /*******************************************************************
    This parameter should be followed by a number n.  If the print_tests
    parameter is included, this parameter is ignored.  If not, this
    parameter causes an announcement about every nth system call test to
    be output to stdout.  This allows the user to keep track of the
    tester's progress when a large number of tests are being done.
    *******************************************************************/
      AnnounceTestsEach = atoi(&argv[++n][0]);
#     ifdef ECHO_OPTIONS
        printf("  Announcing after each %d system calls.\n",
	       AnnounceTestsEach);
#     endif
      AnnounceTestsFlag = TRUE;

    } else if (!strcmp(&argv[n][1], "signal_parent")) {
    /*******************************************************************
    This parameter will cause the fuzz tool to send a SIGUSR1 to its
    parent upon successful completion of all tests.  This argument 
    should only be used by the runmany program.
    *******************************************************************/
#     ifdef ECHO_OPTIONS
        printf("  Will send SIGUSR1 to parent.\n");
#     endif
      SignalParent = TRUE;

    } else {
    /*******************************************************************
    Unrecognized option.
    *******************************************************************/
      printf("\n");
      fprintf(stderr, "Error:  Unrecognized option '%s'.\n", &argv[n][1]);
      usage(argv[0]);
      exit(1);
    } /* if */
  } /* for */
}

void usage( const char* MyName ) {
  fprintf(stderr, "Usage: %s [options]\n", MyName);
  fprintf(stderr, " Supported options: (defaults are enclosed in parenthesis)\n");
  fprintf(stderr, "  -callfile <input file name> (\"in\")\n");
  fprintf(stderr, "  -tests <number of system calls to make> (1000)\n");
  fprintf(stderr, "  -max_parms <number of parameters to use per call> (8)\n");
  fprintf(stderr, "  -logfile <name of log file> (no logging)\n");
  fprintf(stderr, "  -summaryfile <name of summary file> (no summary)\n");
  fprintf(stderr, "  -seed <random number seed> (selected based on time of day)\n");
  fprintf(stderr, "  -random_state <x> <y> <z> [allows more variation of seeds]\n");
  fprintf(stderr, "  -aggressive [turns on aggressive mode] (off)\n");
  fprintf(stderr, "  -catch_sigsegv [installs a SIGSEGV signal handler] (off)\n");
  fprintf(stderr, "  -print_tests [info is printed to the screen about each call] (off)\n");
  fprintf(stderr, "  -announce_tests <n> [Print out a message every n calls] (off)\n");
  fprintf(stderr, "  -signal_parent [send SIGUSR1 to parent at finish] (off)\n");
  fflush(stderr);
}

void process_options(void) {
/***********************************************************************
Set up everything based on the options provided.
***********************************************************************/

  /* Read the syscall definition file. */
  get_call_definitions();

  /* Set up the logging file. */
  if (LoggingFlag) set_up_logging();

  /* Write the summary file. */
  if (SummaryFlag) write_summary();

  /* Initialize the random number generator. */
  seed48(Rand48State);

  /* Set up a signal handler to catch SIGSEGV. */
  if (CatchSIGSEGVFlag) {
    printf("Call to signal() returned %d.",
           (int)signal(SIGSEGV, handle_signal));
    printf("  errno = %d.\n", errno);
    signal(SIGSEGV, SIG_IGN);
  }
  /* Ignore SIGALRM, so that calling alarm() won't exit the tester. */  
  signal(SIGALRM, SIG_IGN);
}


void handle_signal(int sig) {
/***********************************************************************
This signal handler ignores a signal.  Unfortunately, it does not output
the message that it should.  Find out what is wrong here.
***********************************************************************/
  /*_exit(321);*/
  fprintf(stderr, "Just got the signal %d.\n", sig);
  fflush(stderr);
  signal(SIGSEGV, handle_signal);
}


void get_call_definitions(void) {
/***********************************************************************
Read the call definitions from the file that appears as a command line
option or defaults to stdin.
Any line starting with "#" is considered to be a comment and is ignored.
***********************************************************************/
  char ch;
  int tryNextLineFlag;
# ifdef ECHO_SYSCALLS
    int n;
# endif
    
  /* Open the file. */
  assert(CallFileName);
  CallFile = fopen(CallFileName, "r");
  assert(CallFile);

  /* Read the definitions. */
  NSysCall = 0;
  while(!feof(CallFile)) {
    /* Ignore comments and white space. */
    do {
      tryNextLineFlag = FALSE;

      /* Flush whitespace. */
      while (((ch = getc(CallFile)) == ' ') || (ch == '\t') || (ch == '\n'));
      if (feof(CallFile)) break;

      /* Check to see if the line is a comment and flush the line if it
      was.  Afterwards, put the character back and check for another
      comment. */
      if (ch == '#') {
        while (getc(CallFile) != '\n');
	tryNextLineFlag = TRUE;
      } else {
        ungetc(ch, CallFile);
      }
    } while (tryNextLineFlag);

    if (feof(CallFile)) break;

    fscanf(CallFile, "%d %s", &SysCall[NSysCall].num,
                              SysCall[NSysCall].name);
    
    /* Read in the parameter flags. */
    read_parameter_flags();

#   ifdef ECHO_SYSCALLS
      if (NSysCall == 0) printf("\nLoading syscall file:\n");
      printf("  Syscall %3d with parameters '", SysCall[NSysCall].num);
      for (n = 0; n < MAX_PARAMETERS; n++) {
	printf("%d", SysCall[NSysCall].parmFlag[n]);
      }
      printf("' is '%s'\n", SysCall[NSysCall].name);
#   endif

    NSysCall++;
  }

# ifdef PRINT_NUMBER_OF_SYSCALLS_FOUND
    printf("\n%d system calls loaded.\n", NSysCall);
# endif

  /* Close the call file. */
  if (CallFile && CallFile != stdin) fclose(CallFile);
  if (NSysCall == 0) {
    printf("\nNo system calls found--exiting.\n");
    exit(99);
  }
}
    
void read_parameter_flags(void) {
/***********************************************************************
Read in the parameter flags.  There may be any number of them, and they
are either a "1" or a "0".  Read them until a comment or end-of-line is
found.

The parameter flags are:  "0" stored as 0 meaning "not a pointer".
                          "1" stored as 1 meaning "a pointer".

If none appear on the line, none will be read.  Any parameter flag that
is not found is assumed to be set to 0.
***********************************************************************/
  int parm;
  char ch;

  /* Set defaults. */
  for (parm = 0; parm < MAX_PARAMETERS; parm++) {
    SysCall[NSysCall].parmFlag[parm] = 0;
  }

  /* Grab the parameters. */
  parm = 0;
  while ((ch = getc(CallFile)) != '\n') {

    /* If it is a comment, just put it back and the line will get
    flushed. */
    if (ch == '#') {
      ungetc(ch, CallFile);
      break;
    }

    if (parm < MAX_PARAMETERS) {
      if (ch == '0') {
        SysCall[NSysCall].parmFlag[parm++] = 0;
      } else if (ch == '1') {
        SysCall[NSysCall].parmFlag[parm++] = 1;
      }
    }
  }
}


void set_up_logging(void) {
/***********************************************************************
Open the logging file.
***********************************************************************/

  assert(LoggingFlag);
  assert(LogFileName);

  /* Open the file. */
  LogFile = fopen(LogFileName, "w");
  assert(LogFile);

  /* Dump the option information to the log file. */
  fprintf(LogFile, "Using call file '%s'.\n", CallFileName);
  if (SummaryFlag) {
    fprintf(LogFile, "Using summary file '%s'.\n", SummaryFileName);
  }
  fprintf(LogFile, "Performing %d tests.\n", NTest);
  fprintf(LogFile, "The maximum parameters allowed for a call is %d.\n",
          MaxParameters);
  fprintf(LogFile, "The initial random number seed is:  %d %d %d\n",
          Rand48State[0], Rand48State[1], Rand48State[2]);

  fprintf(LogFile, "\n");
}


void write_summary(void) {
/***********************************************************************
Append the information about this run to the summary file.  Close it and
sync right away so that it is not lost.

A timestamp could be added here.
***********************************************************************/

  assert(SummaryFlag);
  assert(SummaryFileName);

  /* Open the file. */
  SummaryFile = fopen(SummaryFileName, "a");
  assert(SummaryFile);

  /* Dump the option information to the summary file. */
  fprintf(SummaryFile, "\n");
  fprintf(SummaryFile, "Using call file '%s'.\n", CallFileName);
  if (LoggingFlag) {
    fprintf(SummaryFile, "Using log file '%s'.\n", LogFileName);
  }
  fprintf(SummaryFile, "Performing %d tests.\n", NTest);
  fprintf(SummaryFile, "The maximum parameters allowed for a call is %d.\n",
          MaxParameters);
  fprintf(SummaryFile,
          "The initial random number seed is:  %d %d %d\n",
          Rand48State[0], Rand48State[1], Rand48State[2]);

  /* Close and re-open the summary file in the best attempt to ensure
  that the data just written gets to the disk. */
  fflush(SummaryFile);
  /*sync();*/
  fsync(fileno(SummaryFile));
}


unsigned short generate_seed(void) {
/***********************************************************************
Generate one of the three 16-bit values needed to define the state for
drand48().
***********************************************************************/
  unsigned short result;

  struct timeval currentTime;
  struct timezone dummyZone;

  /* This does not appear in Solaris' /usr/include/sys/time.h file when
  standard C is chosen. */

  assert(gettimeofday(&currentTime, &dummyZone) == 0);

  result = currentTime.tv_sec % 10000;

  return result;
}
