#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <assert.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>

#ifndef PATH_MAX
  #define PATH_MAX 1023
#endif

/* These are the possible results of a given case. */
#define HUNG 0
#define EARLY 1
#define NORMAL 2

/* Define this if you want to see debugging output on stderr */
/* #define DEBUG */

/* The maximum number of arguments this program will have to pass to
   the fuzz tool (not including the name of the program itself).  */
#define MAXFUZZARGS 20

/************************************************************************ 
Function prototypes 
************************************************************************/
void get_options(int argc, char* argv[]);
void process_options(void);
void handle_child(int sig);
void handle_usr1(int sig);
void usage(const char* myName);
unsigned short generate_seed(void);

/************************************************************************ 
Global variables 
************************************************************************/
int numcases;
int numtests;
char* infileName;
char* sumfileName="summary.runmany";
char* logfileName;

/* String to hold name of the fuzz binary */
char* fuzzName="fuzz";

/* Array of strings to hold the arguments for the fuzz tool.  Need + 2
   for argv[0], and for a NULL to terminate the list. */
char* fuzzargv[MAXFUZZARGS + 2];

/* The time to sleep to see if the fuzz tester has hung */
unsigned int sleepTime;

/* Flags to maintain the status of the current fuzz tester run. */
int childStatus;
int exitStatus;

/* Flag set if runmany should start up each case in logging mode. */
int runWithLog;

/* String to hold the value of the random number seed for the 
   current test. */
char seedStr[10];

/**********************************************************************
Main program
**********************************************************************/
int main(int argc, char *argv[]) {

  int i, execStatus;
  FILE* sumFile;
  pid_t child;
  unsigned short seed = generate_seed();

  get_options(argc, argv);
  process_options();

  /* Open the summary file. */
  sumFile=fopen(sumfileName, "w");
  assert(sumFile);
  fprintf(sumFile, "******************************************\n");
  fprintf(sumFile, "Using call file '%s'\n", infileName);
  fprintf(sumFile, "Performing %d cases.\n", numcases);
  fprintf(sumFile, "Performing %d tests per case.\n", numtests);
  fprintf(sumFile, "******************************************\n\n");
  fflush(sumFile);
  
  /* Install signal handlers for SIGCHLD, and SIGUSR1. */
  signal(SIGCHLD, handle_child);
  signal(SIGUSR1, handle_usr1);

  /************************************************************************* 
    Loop through the number of cases we do, fork off a child to run fuzz, 
    and see what the results are.  Write the result to the summary file.
  *************************************************************************/
  for(i=1; i<=numcases; i++) {
  
  /* Write initial summary info for this case. */
    fprintf(sumFile, "Begining case %d\n", i);
    fprintf(sumFile, "The initial random number seed is: %d\n", (int)seed);
    fflush(sumFile);
    fsync(fileno(sumFile));

    /* Assume the child will exit early.  If it exits normally, we'll
       be told, and if it hangs, we'll decide that later. */
    childStatus=EARLY;

    /* Record the current value of the seed to the string that is 
       passed to the fuzz program. */
    sprintf(seedStr, "%d", (int)seed);

    child=fork();
    if(child) {    /* parent */
      /* Sleep returns 0 if it the requested time has elapsed, meaning
	 that the child has hung and needs to be killed off.  If we
	 get a SIGCHLD, sleep() will exit with a non-zero return
	 value, leaving childStatus alone. */
      if(! sleep(sleepTime)) {
	childStatus=HUNG;
      }
      switch(childStatus) {
      case HUNG:
#ifdef DEBUG
	fprintf(stderr, "About to kill my child, pid: %d.\n", child);
#endif
	kill(child, SIGKILL); 
	fprintf(sumFile, "+++++ This test hung. +++++\n\n");
	printf("+++++ This test hung. +++++\n\n");
	break;
      case EARLY:
	fprintf(sumFile, "+++++ This test exited prematurely with status: %d. +++++\n\n", exitStatus);
	printf("+++++ This test exited prematurely with status: %d. +++++\n\n", exitStatus);
	break;
      case NORMAL:
	fprintf(sumFile, "----- This test completed successfully. -----\n\n");
	printf("----- This test completed successfully. -----\n\n");
	break;
      } 
    } else {     /* child */
      if(runWithLog) {
	sprintf(logfileName, "log.%d", i);
      }
      execStatus=execv(fuzzName, fuzzargv); 
      if(execStatus == -1) {
	fprintf(stderr, "The execv() function failed with errno: %d.\n", errno);
	return(1);
      }
    } 
    seed++;
    fflush(sumFile);
    fsync(fileno(sumFile));
  } 
  return(0);
}

/************************************************************************
process_options() -- Construct the argument vector for the calls to fuzz, 
  based on the arguments to runmany.  Also, compute the time needed to 
  sleep to see if fuzz hung, based on the number of tests we're trying.
************************************************************************/
void process_options(void) {
  int sleepDevisor, x=0;
  assert(infileName);
  assert(sumfileName);
  assert(numtests);
  assert(numcases);
  /* Compute sleepTime */
  if(!sleepTime) {
    if(runWithLog) {
      sleepDevisor = 5;
    } else {
      sleepDevisor = 100;
    }
    if(numtests <= sleepDevisor) {
      sleepTime=3;
    } else {
      sleepTime=(unsigned int)((numtests/sleepDevisor)+2); 
    }
  }
#ifdef DEBUG
  fprintf(stderr, "sleepTime is: %d\n", sleepTime);
#endif
  /* Set up argument array for calls to fuzz */
  fuzzargv[x++]=fuzzName;
  fuzzargv[x++]="-callfile";
  fuzzargv[x++]=infileName;
  fuzzargv[x++]="-tests";
  fuzzargv[x]=malloc(sizeof(numtests) + 1);
  sprintf(fuzzargv[x++], "%d", numtests);
  fuzzargv[x++]="-signal_parent";
  /*
  fuzzargv[x++]="-summaryfile";
  fuzzargv[x++]="sum";
  */
  fuzzargv[x++]="-random_state";
  fuzzargv[x++]=seedStr;
  fuzzargv[x++]=seedStr;
  fuzzargv[x++]=seedStr;
  if(runWithLog) {
    logfileName=malloc(PATH_MAX);
    fuzzargv[x++]="-logfile";
    fuzzargv[x++]=logfileName;
  } 
  fuzzargv[x]=NULL;
}

/************************************************************************
handle_child() -- signal handler for SIGCHLD, which runmany recieves
  whenever its child terminates.  
************************************************************************/
void handle_child(int sig) {
  int status;
  pid_t child;
#ifdef DEBUG
  fprintf(stderr, "Just got SIGCHLD.\n");
  fflush(stderr);
#endif
  child=wait(&status);
  exitStatus=WEXITSTATUS(status);

#ifdef DEBUG
  fprintf(stderr, "Process %d exited with status: %d.\n", (int)child,
	  exitStatus);
  fflush(stderr);
#endif
  signal(SIGCHLD, handle_child);
}

/************************************************************************
handle_usr1() -- signal handler for SIGUSR1, which runmany recieves
  whenever the fuzz program exits after completing all its tests.  
************************************************************************/
void handle_usr1(int sig) {
#ifdef DEBUG
  fprintf(stderr, "Just got SIGUSR1.\n");
  fflush(stderr);
#endif
  childStatus=NORMAL;
  signal(SIGUSR1, handle_usr1);
}

/************************************************************************
get_options() -- Parse the command line arguments and store the
  results in global variables.
************************************************************************/
void get_options(int argc, char* argv[]) {
  int n;
  char* myName=argv[0];

  /* Default values. */
  numcases=100;
  numtests=1000;
  runWithLog=0;
  sleepTime=0;   /* This will be changed later if it's still 0 */
  
  for (n=1; n<argc; n++) {
    if (argv[n][0] != '-') {
      fprintf(stderr, "Error:  Argument '%s' is not valid.\n", argv[n]);
      usage(myName);
      exit(1);
    }
    if (!strcmp(&argv[n][1], "cases")) {
      numcases=atoi(&argv[++n][0]);

    } else if (!strcmp(&argv[n][1], "tests")) {
      numtests=atoi(&argv[++n][0]);

    } else if (!strcmp(&argv[n][1], "callfile")) {
      infileName=argv[++n];

    } else if (!strcmp(&argv[n][1], "sleep")) {
      sleepTime=atoi(&argv[++n][0]);

    } else if (!strcmp(&argv[n][1], "log")) {
      runWithLog=1;

    } else {
      fprintf(stderr, "Error:  Unrecognized option '%s'.\n", &argv[n][1]);
      usage(myName);
      exit(1);
    }
  }
  if(!infileName) {
    infileName=strdup("in");
  }
}

/************************************************************************
usage() -- Print out usage information for this program.
************************************************************************/
void usage(const char* myName) {
  fprintf(stderr, "Usage: %s [options]\n", myName);
  fprintf(stderr, "  Available options: (defaults are listed)\n");
  fprintf(stderr, "  -cases <number of separate cases to run> (100)\n");
  fprintf(stderr, "  -tests <number of calls to do per case> (1000)\n");
  fprintf(stderr, "  -callfile <name of file to use for fuzz input file> ('in')\n");
  fprintf(stderr, "  -log (run each case with logging on -- defaults to no)\n");
  fprintf(stderr, "  -sleep <number of seconds for timeout of hung system calls>\n");
  fprintf(stderr, "         (this option is not recomended, since runmany\n");
  fprintf(stderr, "         figures out a reasonable value automatically)\n");
}

unsigned short generate_seed(void) {
/***********************************************************************
Generate one of the three 16-bit values needed to define the state for
drand48().
***********************************************************************/
  struct timeval currentTime;
  struct timezone dummyZone;
  assert(gettimeofday(&currentTime, &dummyZone) == 0);
  return(currentTime.tv_sec % 10000);
}
